from uuid import uuid4

from orchestrator.db.models import ProductTable, SubscriptionTable
# noinspection PyProtectedMember
from orchestrator.forms import FormPage
from orchestrator.forms.validators import Choice, UniqueConstrainedList
from orchestrator.targets import Target
from orchestrator.types import FormGenerator, State
from orchestrator.types import SubscriptionLifecycle, UUIDstr
from orchestrator.workflow import done, init, step, workflow
from orchestrator.workflows.steps import resync, set_status
from orchestrator.workflows.steps import store_process_subscription
from orchestrator.workflows.utils import wrap_create_initial_input_form

from gso.products.product_blocks import PhyPortCapacity
from gso.products.product_blocks.iptrunk import IptrunkType
from gso.products.product_types.device import Device
from gso.products.product_types.iptrunk import IptrunkInactive, \
    IptrunkProvisioning
# noinspection PyProtectedMember
from gso.services import provisioning_proxy, _ipam
from gso.services.provisioning_proxy import confirm_pp_results, \
    await_pp_results


def initial_input_form_generator(product_name: str) -> FormGenerator:
    # TODO: we need additional validation:
    # * interface names must be validated

    devices = {}
    for device_id, device_description in (
            SubscriptionTable.query.join(ProductTable)
            .filter(
                ProductTable.product_type == 'Device',
                SubscriptionTable.status == 'active',
            )
            .with_entities(SubscriptionTable.subscription_id,
                           SubscriptionTable.description)
            .all()
    ):
        devices[str(device_id)] = device_description

    class CreateIptrunkForm(FormPage):
        class Config:
            title = product_name

        geant_s_sid: str
        iptrunk_description: str
        iptrunk_type: IptrunkType
        iptrunk_speed: PhyPortCapacity
        iptrunk_minimum_links: int

    initial_user_input = yield CreateIptrunkForm

    class AeMembersList(UniqueConstrainedList[str]):
        min_items = initial_user_input.iptrunk_minimum_links
        max_items = initial_user_input.iptrunk_minimum_links

    DeviceEnumA = Choice('Device A', zip(devices.keys(), devices.items()))

    class CreateIptrunkSideAForm(FormPage):
        class Config:
            title = 'Provide subscription details for side A of the trunk.'

        iptrunk_sideA_node_id: DeviceEnumA
        iptrunk_sideA_ae_iface: str
        iptrunk_sideA_ae_geant_a_sid: str
        iptrunk_sideA_ae_members: AeMembersList
        iptrunk_sideA_ae_members_descriptions: AeMembersList

    user_input_side_a = yield CreateIptrunkSideAForm

    # We remove the selected device for side A, to prevent any loops
    devices.pop(str(user_input_side_a.iptrunk_sideA_node_id.name))
    DeviceEnumB = Choice('Device B', zip(devices.keys(), devices.items()))

    class CreateIptrunkSideBForm(FormPage):
        class Config:
            title = 'Provide subscription details for side B of the trunk.'

        iptrunk_sideB_node_id: DeviceEnumB
        iptrunk_sideB_ae_iface: str
        iptrunk_sideB_ae_geant_a_sid: str
        iptrunk_sideB_ae_members: AeMembersList
        iptrunk_sideB_ae_members_descriptions: AeMembersList

    user_input_side_b = yield CreateIptrunkSideBForm

    return initial_user_input.dict() | \
        user_input_side_a.dict() | \
        user_input_side_b.dict()


@step('Create subscription')
def create_subscription(product: UUIDstr) -> State:
    subscription = IptrunkInactive.from_product_id(product, uuid4())

    return {
        'subscription': subscription,
        'subscription_id': subscription.subscription_id,
    }


@step('Get information from IPAM')
def get_info_from_ipam(subscription: IptrunkProvisioning) -> State:
    subscription.iptrunk.iptrunk_ipv4_network \
        = _ipam.allocate_service_ipv4_network(
            service_type='TRUNK',
            comment=subscription.iptrunk.iptrunk_description
        ).v4
    subscription.iptrunk.iptrunk_ipv6_network \
        = _ipam.allocate_service_ipv6_network(
            service_type='TRUNK',
            comment=subscription.iptrunk.iptrunk_description
        ).v6
    return {'subscription': subscription}


@step('Initialize subscription')
def initialize_subscription(
        subscription: IptrunkInactive,
        geant_s_sid: str,
        iptrunk_type: IptrunkType,
        iptrunk_description: str,
        iptrunk_speed: str,
        iptrunk_minimum_links: int,
        iptrunk_sideA_node_id: str,
        iptrunk_sideA_ae_iface: str,
        iptrunk_sideA_ae_geant_a_sid: str,
        iptrunk_sideA_ae_members: list[str],
        iptrunk_sideA_ae_members_descriptions: list[str],
        iptrunk_sideB_node_id: str,
        iptrunk_sideB_ae_iface: str,
        iptrunk_sideB_ae_geant_a_sid: str,
        iptrunk_sideB_ae_members: list[str],
        iptrunk_sideB_ae_members_descriptions: list[str]
) -> State:
    subscription.iptrunk.geant_s_sid = geant_s_sid
    subscription.iptrunk.iptrunk_description = iptrunk_description
    subscription.iptrunk.iptrunk_type = iptrunk_type
    subscription.iptrunk.iptrunk_speed = iptrunk_speed
    subscription.iptrunk.iptrunk_isis_metric = 9000
    subscription.iptrunk.iptrunk_minimum_links = iptrunk_minimum_links

    subscription.iptrunk.iptrunk_sideA_node = \
        Device.from_subscription(iptrunk_sideA_node_id).device
    subscription.iptrunk.iptrunk_sideA_ae_iface = iptrunk_sideA_ae_iface
    subscription.iptrunk.iptrunk_sideA_ae_geant_a_sid \
        = iptrunk_sideA_ae_geant_a_sid
    subscription.iptrunk.iptrunk_sideA_ae_members = iptrunk_sideA_ae_members
    subscription.iptrunk.iptrunk_sideA_ae_members_description \
        = iptrunk_sideA_ae_members_descriptions

    subscription.iptrunk.iptrunk_sideB_node = \
        Device.from_subscription(iptrunk_sideB_node_id).device
    subscription.iptrunk.iptrunk_sideB_ae_iface \
        = iptrunk_sideB_ae_iface
    subscription.iptrunk.iptrunk_sideB_ae_geant_a_sid \
        = iptrunk_sideB_ae_geant_a_sid
    subscription.iptrunk.iptrunk_sideB_ae_members = iptrunk_sideB_ae_members
    subscription.iptrunk.iptrunk_sideB_ae_members_description \
        = iptrunk_sideB_ae_members_descriptions

    subscription.description = f'IP trunk, geant_s_sid:{geant_s_sid}'
    subscription = IptrunkProvisioning.from_other_lifecycle(
        subscription, SubscriptionLifecycle.PROVISIONING
    )

    return {'subscription': subscription}


@step('Provision IP trunk interface [DRY RUN]')
def provision_ip_trunk_iface_dry(subscription: IptrunkProvisioning,
                                 process_id: UUIDstr) -> State:
    provisioning_proxy.provision_ip_trunk(subscription, process_id,
                                          'trunk_interface')

    return {'subscription': subscription,
            'label_text': 'This is a dry run for the deployment of a new IP '
                          'trunk. Deployment is being taken care of by the '
                          'provisioning proxy, please wait for the results to '
                          'come back before continuing.'
            }


@step('Provision IP trunk interface [FOR REAL]')
def provision_ip_trunk_iface_real(subscription: IptrunkProvisioning,
                                  process_id: UUIDstr) -> State:
    provisioning_proxy.provision_ip_trunk(subscription, process_id,
                                          'trunk_interface', False)

    return {'subscription': subscription,
            'label_text': 'This is a live deployment of a new IP trunk. '
                          'Deployment is being taken care of by the '
                          'provisioning proxy, please wait for the results to '
                          'come back before continuing.'
            }


@step('Provision IP trunk ISIS interface [DRY RUN]')
def provision_ip_trunk_isis_iface_dry(subscription: IptrunkProvisioning,
                                      process_id: UUIDstr) -> State:
    provisioning_proxy.provision_ip_trunk(subscription, process_id,
                                          'isis_interface')

    return {'subscription': subscription,
            'label_text': 'This is a dry run for the deployment of a new IP '
                          'trunk ISIS interface. '
                          'Deployment is being taken care of by the '
                          'provisioning proxy, please wait for the results to '
                          'come back before continuing.'
            }


@step('Provision IP trunk ISIS interface [FOR REAL]')
def provision_ip_trunk_isis_iface_real(subscription: IptrunkProvisioning,
                                       process_id: UUIDstr) -> State:
    provisioning_proxy.provision_ip_trunk(subscription, process_id,
                                          'isis_interface', False)

    return {'subscription': subscription,
            'label_text': 'This is a live deployment of a new IP trunk '
                          'ISIS interface. '
                          'Deployment is being taken care of by the '
                          'provisioning proxy, please wait for the results to '
                          'come back before continuing.'
            }


@step('Provision IP trunk LDP interface [DRY RUN]')
def provision_ip_trunk_ldp_iface_dry(subscription: IptrunkProvisioning,
                                     process_id: UUIDstr) -> State:
    provisioning_proxy.provision_ip_trunk(subscription, process_id,
                                          'ldp_interface')

    return {'subscription': subscription,
            'label_text': 'This is a dry run for the deployment of a new IP '
                          'trunk LDP interface. '
                          'Deployment is being taken care of by the '
                          'provisioning proxy, please wait for the results to '
                          'come back before continuing.'
            }


@step('Provision IP trunk LDP interface [FOR REAL]')
def provision_ip_trunk_ldp_iface_real(subscription: IptrunkProvisioning,
                                      process_id: UUIDstr) -> State:
    provisioning_proxy.provision_ip_trunk(subscription, process_id,
                                          'ldp_interface', False)

    return {'subscription': subscription,
            'label_text': 'This is a live deployment of a new IP trunk '
                          'LDP interface. '
                          'Deployment is being taken care of by the '
                          'provisioning proxy, please wait for the results to '
                          'come back before continuing.'
            }


@step('Provision IP trunk LLDP interface [DRY RUN]')
def provision_ip_trunk_lldp_iface_dry(subscription: IptrunkProvisioning,
                                      process_id: UUIDstr) -> State:
    provisioning_proxy.provision_ip_trunk(subscription, process_id,
                                          'lldp_interface')

    return {'subscription': subscription,
            'label_text': 'This is a dry run for the deployment of a new IP '
                          'trunk LLDP interface. '
                          'Deployment is being taken care of by the '
                          'provisioning proxy, please wait for the results to '
                          'come back before continuing.'
            }


@step('Provision IP trunk LLDP interface [FOR REAL]')
def provision_ip_trunk_lldp_iface_real(subscription: IptrunkProvisioning,
                                       process_id: UUIDstr) -> State:
    provisioning_proxy.provision_ip_trunk(subscription, process_id,
                                          'lldp_interface', False)

    return {'subscription': subscription,
            'label_text': 'This is a live deployment of a new IP trunk '
                          'LLDP interface. '
                          'Deployment is being taken care of by the '
                          'provisioning proxy, please wait for the results to '
                          'come back before continuing.'
            }


@workflow(
    'Create IP trunk',
    initial_input_form=wrap_create_initial_input_form(
        initial_input_form_generator),
    target=Target.CREATE,
)
def create_iptrunk():
    return (
            init
            >> create_subscription
            >> store_process_subscription(Target.CREATE)
            >> initialize_subscription
            >> get_info_from_ipam
            >> provision_ip_trunk_iface_dry
            >> await_pp_results
            >> confirm_pp_results
            >> provision_ip_trunk_iface_real
            >> await_pp_results
            >> confirm_pp_results
            >> provision_ip_trunk_isis_iface_dry
            >> await_pp_results
            >> confirm_pp_results
            >> provision_ip_trunk_isis_iface_real
            >> await_pp_results
            >> confirm_pp_results
            >> provision_ip_trunk_ldp_iface_dry
            >> await_pp_results
            >> confirm_pp_results
            >> provision_ip_trunk_ldp_iface_real
            >> await_pp_results
            >> confirm_pp_results
            >> provision_ip_trunk_lldp_iface_dry
            >> await_pp_results
            >> confirm_pp_results
            >> provision_ip_trunk_lldp_iface_real
            >> await_pp_results
            >> confirm_pp_results
            >> set_status(SubscriptionLifecycle.ACTIVE)
            >> resync
            >> done
    )
